@@ -1,6 +1,6 @@
 # Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
-# For details: https://github.com/PyCQA/astroid/blob/main/LICENSE
-# Copyright (c) https://github.com/PyCQA/astroid/blob/main/CONTRIBUTORS.txt
+# For details: https://github.com/pylint-dev/astroid/blob/main/LICENSE
+# Copyright (c) https://github.com/pylint-dev/astroid/blob/main/CONTRIBUTORS.txt
 
 """
 Data object model, as per https://docs.python.org/3/reference/datamodel.html.
@@ -27,8 +27,9 @@ import itertools
 import os
 import pprint
 import types
+from collections.abc import Iterator
 from functools import lru_cache
-from typing import TYPE_CHECKING, Any
+from typing import TYPE_CHECKING, Any, Literal
 
 import astroid
 from astroid import bases, nodes, util
@@ -36,12 +37,9 @@ from astroid.context import InferenceContext, copy_context
 from astroid.exceptions import AttributeInferenceError, InferenceError, NoDefault
 from astroid.manager import AstroidManager
 from astroid.nodes import node_classes
-
-objects = util.lazy_import("objects")
-builder = util.lazy_import("builder")
+from astroid.typing import InferenceResult, SuccessfulInferenceResult
 
 if TYPE_CHECKING:
-    from astroid import builder
     from astroid.objects import Property
 
 IMPL_PREFIX = "attr_"
@@ -49,7 +47,13 @@ LEN_OF_IMPL_PREFIX = len(IMPL_PREFIX)
 
 
 def _dunder_dict(instance, attributes):
-    obj = node_classes.Dict(parent=instance)
+    obj = node_classes.Dict(
+        parent=instance,
+        lineno=instance.lineno,
+        col_offset=instance.col_offset,
+        end_lineno=instance.end_lineno,
+        end_col_offset=instance.end_col_offset,
+    )
 
     # Convert the keys to node strings
     keys = [
@@ -110,16 +114,16 @@ class ObjectModel:
         # underlying data model and to the instance for which it got accessed.
         return self(instance)
 
-    def __contains__(self, name):
+    def __contains__(self, name) -> bool:
         return name in self.attributes()
 
-    @lru_cache()  # noqa
+    @lru_cache  # noqa
     def attributes(self) -> list[str]:
         """Get the attributes which are exported by this object model."""
         return [o[LEN_OF_IMPL_PREFIX:] for o in dir(self) if o.startswith(IMPL_PREFIX)]
 
     def lookup(self, name):
-        """Look up the given *name* in the current model
+        """Look up the given *name* in the current model.
 
         It should return an AST or an interpreter object,
         but if the name is not found, then an AttributeInferenceError will be raised.
@@ -131,18 +135,22 @@ class ObjectModel:
     @property
     def attr___new__(self) -> bases.BoundMethod:
         """Calling cls.__new__(type) on an object returns an instance of 'type'."""
+        from astroid import builder  # pylint: disable=import-outside-toplevel
+
         node: nodes.FunctionDef = builder.extract_node(
             """def __new__(self, cls): return cls()"""
         )
         # We set the parent as being the ClassDef of 'object' as that
         # triggers correct inference as a call to __new__ in bases.py
-        node.parent: nodes.ClassDef = AstroidManager().builtins_module["object"]
+        node.parent = AstroidManager().builtins_module["object"]
 
         return bases.BoundMethod(proxy=node, bound=_get_bound_node(self))
 
     @property
     def attr___init__(self) -> bases.BoundMethod:
         """Calling cls.__init__() normally returns None."""
+        from astroid import builder  # pylint: disable=import-outside-toplevel
+
         # The *args and **kwargs are necessary not to trigger warnings about missing
         # or extra parameters for '__init__' methods we don't infer correctly.
         # This BoundMethod is the fallback value for those.
@@ -151,7 +159,7 @@ class ObjectModel:
         )
         # We set the parent as being the ClassDef of 'object' as that
         # is where this method originally comes from
-        node.parent: nodes.ClassDef = AstroidManager().builtins_module["object"]
+        node.parent = AstroidManager().builtins_module["object"]
 
         return bases.BoundMethod(proxy=node, bound=_get_bound_node(self))
 
@@ -261,7 +269,13 @@ class FunctionModel(ObjectModel):
 
     @property
     def attr___annotations__(self):
-        obj = node_classes.Dict(parent=self._instance)
+        obj = node_classes.Dict(
+            parent=self._instance,
+            lineno=self._instance.lineno,
+            col_offset=self._instance.col_offset,
+            end_lineno=self._instance.end_lineno,
+            end_col_offset=self._instance.end_col_offset,
+        )
 
         if not self._instance.returns:
             returns = None
@@ -295,7 +309,13 @@ class FunctionModel(ObjectModel):
 
     @property
     def attr___dict__(self):
-        return node_classes.Dict(parent=self._instance)
+        return node_classes.Dict(
+            parent=self._instance,
+            lineno=self._instance.lineno,
+            col_offset=self._instance.col_offset,
+            end_lineno=self._instance.end_lineno,
+            end_col_offset=self._instance.end_col_offset,
+        )
 
     attr___globals__ = attr___dict__
 
@@ -312,7 +332,13 @@ class FunctionModel(ObjectModel):
                 yield name, default
 
         args = self._instance.args
-        obj = node_classes.Dict(parent=self._instance)
+        obj = node_classes.Dict(
+            parent=self._instance,
+            lineno=self._instance.lineno,
+            col_offset=self._instance.col_offset,
+            end_lineno=self._instance.end_lineno,
+            end_col_offset=self._instance.end_col_offset,
+        )
         defaults = dict(_default_args(args, obj))
 
         obj.postinit(list(defaults.items()))
@@ -327,14 +353,20 @@ class FunctionModel(ObjectModel):
         func = self._instance
 
         class DescriptorBoundMethod(bases.BoundMethod):
-            """Bound method which knows how to understand calling descriptor binding."""
+            """Bound method which knows how to understand calling descriptor
+            binding.
+            """
 
-            def implicit_parameters(self):
+            def implicit_parameters(self) -> Literal[0]:
                 # Different than BoundMethod since the signature
                 # is different.
                 return 0
 
-            def infer_call_result(self, caller, context=None):
+            def infer_call_result(
+                self,
+                caller: SuccessfulInferenceResult | None,
+                context: InferenceContext | None = None,
+            ) -> Iterator[bases.BoundMethod]:
                 if len(caller.args) > 2 or len(caller.args) < 1:
                     raise InferenceError(
                         "Invalid arguments for descriptor binding",
@@ -348,7 +380,7 @@ class FunctionModel(ObjectModel):
                 except StopIteration as e:
                     raise InferenceError(context=context, node=caller.args[0]) from e
 
-                if cls is astroid.Uninferable:
+                if isinstance(cls, util.UninferableBase):
                     raise InferenceError(
                         "Invalid class inferred", target=self, context=context
                     )
@@ -366,6 +398,8 @@ class FunctionModel(ObjectModel):
                     lineno=func.lineno,
                     col_offset=func.col_offset,
                     parent=func.parent,
+                    end_lineno=func.end_lineno,
+                    end_col_offset=func.end_col_offset,
                 )
                 # pylint: disable=no-member
                 new_func.postinit(
@@ -382,7 +416,7 @@ class FunctionModel(ObjectModel):
 
             @property
             def args(self):
-                """Overwrite the underlying args to match those of the underlying func
+                """Overwrite the underlying args to match those of the underlying func.
 
                 Usually the underlying *func* is a function/method, as in:
 
@@ -393,12 +427,24 @@ class FunctionModel(ObjectModel):
                 we get a new object which has two parameters, *self* and *type*.
                 """
                 nonlocal func
+                arguments = astroid.Arguments(
+                    parent=func.args.parent, vararg=None, kwarg=None
+                )
+
                 positional_or_keyword_params = func.args.args.copy()
-                positional_or_keyword_params.append(astroid.AssignName(name="type"))
+                positional_or_keyword_params.append(
+                    astroid.AssignName(
+                        name="type",
+                        lineno=0,
+                        col_offset=0,
+                        parent=arguments,
+                        end_lineno=None,
+                        end_col_offset=None,
+                    )
+                )
 
                 positional_only_params = func.args.posonlyargs.copy()
 
-                arguments = astroid.Arguments(parent=func.args.parent)
                 arguments.postinit(
                     args=positional_or_keyword_params,
                     posonlyargs=positional_only_params,
@@ -406,6 +452,8 @@ class FunctionModel(ObjectModel):
                     kwonlyargs=[],
                     kw_defaults=[],
                     annotations=[],
+                    kwonlyargs_annotations=[],
+                    posonlyargs_annotations=[],
                 )
                 return arguments
 
@@ -480,7 +528,11 @@ class ClassModel(ObjectModel):
         # Cls.mro is a method and we need to return one in order to have a proper inference.
         # The method we're returning is capable of inferring the underlying MRO though.
         class MroBoundMethod(bases.BoundMethod):
-            def infer_call_result(self, caller, context=None):
+            def infer_call_result(
+                self,
+                caller: SuccessfulInferenceResult | None,
+                context: InferenceContext | None = None,
+            ) -> Iterator[node_classes.Tuple]:
                 yield other_self.attr___mro__
 
         implicit_metaclass = self._instance.implicit_metaclass()
@@ -504,7 +556,7 @@ class ClassModel(ObjectModel):
 
     @property
     def attr___subclasses__(self):
-        """Get the subclasses of the underlying class
+        """Get the subclasses of the underlying class.
 
         This looks only in the current module for retrieving the subclasses,
         thus it might miss a couple of them.
@@ -526,7 +578,11 @@ class ClassModel(ObjectModel):
         obj.postinit(classes)
 
         class SubclassesBoundMethod(bases.BoundMethod):
-            def infer_call_result(self, caller, context=None):
+            def infer_call_result(
+                self,
+                caller: SuccessfulInferenceResult | None,
+                context: InferenceContext | None = None,
+            ) -> Iterator[node_classes.List]:
                 yield obj
 
         implicit_metaclass = self._instance.implicit_metaclass()
@@ -535,7 +591,13 @@ class ClassModel(ObjectModel):
 
     @property
     def attr___dict__(self):
-        return node_classes.Dict(parent=self._instance)
+        return node_classes.Dict(
+            parent=self._instance,
+            lineno=self._instance.lineno,
+            col_offset=self._instance.col_offset,
+            end_lineno=self._instance.end_lineno,
+            end_col_offset=self._instance.end_col_offset,
+        )
 
     @property
     def attr___call__(self):
@@ -582,6 +644,52 @@ class UnboundMethodModel(ObjectModel):
     attr_im_self = attr___self__
 
 
+class ContextManagerModel(ObjectModel):
+    """Model for context managers.
+
+    Based on 3.3.9 of the Data Model documentation:
+    https://docs.python.org/3/reference/datamodel.html#with-statement-context-managers
+    """
+
+    @property
+    def attr___enter__(self) -> bases.BoundMethod:
+        """Representation of the base implementation of __enter__.
+
+        As per Python documentation:
+        Enter the runtime context related to this object. The with statement
+        will bind this method's return value to the target(s) specified in the
+        as clause of the statement, if any.
+        """
+        from astroid import builder  # pylint: disable=import-outside-toplevel
+
+        node: nodes.FunctionDef = builder.extract_node("""def __enter__(self): ...""")
+        # We set the parent as being the ClassDef of 'object' as that
+        # is where this method originally comes from
+        node.parent = AstroidManager().builtins_module["object"]
+
+        return bases.BoundMethod(proxy=node, bound=_get_bound_node(self))
+
+    @property
+    def attr___exit__(self) -> bases.BoundMethod:
+        """Representation of the base implementation of __exit__.
+
+        As per Python documentation:
+        Exit the runtime context related to this object. The parameters describe the
+        exception that caused the context to be exited. If the context was exited
+        without an exception, all three arguments will be None.
+        """
+        from astroid import builder  # pylint: disable=import-outside-toplevel
+
+        node: nodes.FunctionDef = builder.extract_node(
+            """def __exit__(self, exc_type, exc_value, traceback): ..."""
+        )
+        # We set the parent as being the ClassDef of 'object' as that
+        # is where this method originally comes from
+        node.parent = AstroidManager().builtins_module["object"]
+
+        return bases.BoundMethod(proxy=node, bound=_get_bound_node(self))
+
+
 class BoundMethodModel(FunctionModel):
     @property
     def attr___func__(self):
@@ -592,7 +700,7 @@ class BoundMethodModel(FunctionModel):
         return self._instance.bound
 
 
-class GeneratorModel(FunctionModel):
+class GeneratorModel(FunctionModel, ContextManagerModel):
     def __new__(cls, *args, **kwargs):
         # Append the values from the GeneratorType unto this object.
         ret = super().__new__(cls, *args, **kwargs)
@@ -665,11 +773,8 @@ class InstanceModel(ObjectModel):
 
 class ExceptionInstanceModel(InstanceModel):
     @property
-    def attr_args(self):
-        message = node_classes.Const("")
-        args = node_classes.Tuple(parent=self._instance)
-        args.postinit((message,))
-        return args
+    def attr_args(self) -> nodes.Tuple:
+        return nodes.Tuple(parent=self._instance)
 
     @property
     def attr___traceback__(self):
@@ -713,7 +818,7 @@ class ImportErrorInstanceModel(ExceptionInstanceModel):
 class UnicodeDecodeErrorInstanceModel(ExceptionInstanceModel):
     @property
     def attr_object(self):
-        return node_classes.Const("")
+        return node_classes.Const(b"")
 
 
 BUILTIN_EXCEPTIONS = {
@@ -749,7 +854,11 @@ class DictModel(ObjectModel):
         """Generate a bound method that can infer the given *obj*."""
 
         class DictMethodBoundMethod(astroid.BoundMethod):
-            def infer_call_result(self, caller, context=None):
+            def infer_call_result(
+                self,
+                caller: SuccessfulInferenceResult | None,
+                context: InferenceContext | None = None,
+            ) -> Iterator[InferenceResult]:
                 yield obj
 
         meth = next(self._instance._proxied.igetattr(name), None)
@@ -757,6 +866,8 @@ class DictModel(ObjectModel):
 
     @property
     def attr_items(self):
+        from astroid import objects  # pylint: disable=import-outside-toplevel
+
         elems = []
         obj = node_classes.List(parent=self._instance)
         for key, value in self._instance.items:
@@ -765,34 +876,46 @@ class DictModel(ObjectModel):
             elems.append(elem)
         obj.postinit(elts=elems)
 
-        obj = objects.DictItems(obj)
-        return self._generic_dict_attribute(obj, "items")
+        items_obj = objects.DictItems(obj)
+        return self._generic_dict_attribute(items_obj, "items")
 
     @property
     def attr_keys(self):
+        from astroid import objects  # pylint: disable=import-outside-toplevel
+
         keys = [key for (key, _) in self._instance.items]
         obj = node_classes.List(parent=self._instance)
         obj.postinit(elts=keys)
 
-        obj = objects.DictKeys(obj)
-        return self._generic_dict_attribute(obj, "keys")
+        keys_obj = objects.DictKeys(obj)
+        return self._generic_dict_attribute(keys_obj, "keys")
 
     @property
     def attr_values(self):
+        from astroid import objects  # pylint: disable=import-outside-toplevel
 
         values = [value for (_, value) in self._instance.items]
         obj = node_classes.List(parent=self._instance)
         obj.postinit(values)
 
-        obj = objects.DictValues(obj)
-        return self._generic_dict_attribute(obj, "values")
+        values_obj = objects.DictValues(obj)
+        return self._generic_dict_attribute(values_obj, "values")
 
 
 class PropertyModel(ObjectModel):
-    """Model for a builtin property"""
+    """Model for a builtin property."""
 
     def _init_function(self, name):
-        args = nodes.Arguments()
+        function = nodes.FunctionDef(
+            name=name,
+            parent=self._instance,
+            lineno=self._instance.lineno,
+            col_offset=self._instance.col_offset,
+            end_lineno=self._instance.end_lineno,
+            end_col_offset=self._instance.end_col_offset,
+        )
+
+        args = nodes.Arguments(parent=function, vararg=None, kwarg=None)
         args.postinit(
             args=[],
             defaults=[],
@@ -804,8 +927,6 @@ class PropertyModel(ObjectModel):
             kwonlyargs_annotations=[],
         )
 
-        function = nodes.FunctionDef(name=name, parent=self._instance)
-
         function.postinit(args=args, body=[])
         return function
 
@@ -814,7 +935,11 @@ class PropertyModel(ObjectModel):
         func = self._instance
 
         class PropertyFuncAccessor(nodes.FunctionDef):
-            def infer_call_result(self, caller=None, context=None):
+            def infer_call_result(
+                self,
+                caller: SuccessfulInferenceResult | None,
+                context: InferenceContext | None = None,
+            ) -> Iterator[InferenceResult]:
                 nonlocal func
                 if caller and len(caller.args) != 1:
                     raise InferenceError(
@@ -825,7 +950,14 @@ class PropertyModel(ObjectModel):
                     caller=caller, context=context
                 )
 
-        property_accessor = PropertyFuncAccessor(name="fget", parent=self._instance)
+        property_accessor = PropertyFuncAccessor(
+            name="fget",
+            parent=self._instance,
+            lineno=self._instance.lineno,
+            col_offset=self._instance.col_offset,
+            end_lineno=self._instance.end_lineno,
+            end_col_offset=self._instance.end_col_offset,
+        )
         property_accessor.postinit(args=func.args, body=func.body)
         return property_accessor
 
@@ -855,7 +987,11 @@ class PropertyModel(ObjectModel):
             )
 
         class PropertyFuncAccessor(nodes.FunctionDef):
-            def infer_call_result(self, caller=None, context=None):
+            def infer_call_result(
+                self,
+                caller: SuccessfulInferenceResult | None,
+                context: InferenceContext | None = None,
+            ) -> Iterator[InferenceResult]:
                 nonlocal func_setter
                 if caller and len(caller.args) != 2:
                     raise InferenceError(
@@ -863,7 +999,14 @@ class PropertyModel(ObjectModel):
                     )
                 yield from func_setter.infer_call_result(caller=caller, context=context)
 
-        property_accessor = PropertyFuncAccessor(name="fset", parent=self._instance)
+        property_accessor = PropertyFuncAccessor(
+            name="fset",
+            parent=self._instance,
+            lineno=self._instance.lineno,
+            col_offset=self._instance.col_offset,
+            end_lineno=self._instance.end_lineno,
+            end_col_offset=self._instance.end_col_offset,
+        )
         property_accessor.postinit(args=func_setter.args, body=func_setter.body)
         return property_accessor
 
